/*  plumb - the pipe shell - administration of plumb timers
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <libporty/host/time.h>
#include <assert.h>
#include "process.h"
#include "vp.h"
#include "timer.h"
#include "config.h"

/* how many seconds to wait in poll when no timer runs - could be infinite */
#define DEFAULT_PERIOD 60.0

double now = 0.0;               /* current time */
static double next = 0.0;       /* earliest time any known timer triggers */

typedef struct tmr_s tmr_t;

struct tmr_s {
	proc_virt_t *proc;
	double when;
	void *data;
	int del;
	tmr_t *prev, *next;
};

tmr_t *timers = NULL;

/* inline is not portable, make sure this gets "inlined" */
#define update_next(when) \
	do { \
		if (when < next) \
			next = when; \
	} while(0)


int timer_add_delta(proc_virt_t *proc, double delta, void *data)
{
	tmr_t *t;

	assert(proc != NULL);

	if (delta < 0)
		return -1;

/* 0 means "next iteration"; our timing is in ms in poll, next is anything
   within smaller than that */
	if (delta == 0)
		delta = 0.0001;

	now = P_dtime();
	t = malloc(sizeof(tmr_t));
	t->proc = proc;
	t->when = now + delta;
	t->data = data;
	t->prev = NULL;
	t->del  = 0;
	t->next = timers;
	if (timers != NULL)
		timers->prev = t;
	timers = t;
	update_next(t->when);
	return 0;
}

#ifdef TIMER_DEBUG
#define timer_print timer_print_
static void timer_print_(char *prefix)
{
	tmr_t *t;
	fprintf(stderr, "%s\n", prefix);
	for(t = timers; t != NULL; t = t->next) {
		fprintf(stderr, "  %f %p\n", t->when, t->proc);
	}
}
#else
#define timer_print(x)
#endif

static void timer_del(tmr_t *t, int mark_only)
{
	if (mark_only) {
		t->del = 1;
		return;
	}

	timer_print("before del:");
	/* unlink and free */
	if (t->prev == NULL)
		timers = t->next;
	else
		t->prev->next = t->next;
	if (t->next != NULL)
		t->next->prev = t->prev;
	free(t);
	timer_print("after del:");
}

int timer_del_proc(proc_virt_t *proc)
{
	tmr_t *t, *tnext;
	for(t = timers; t != NULL; t = tnext) {
		tnext = t->next;
		if (t->proc == proc)
			timer_del(t, 1);
	}
	return 0;
}

int timer_run(void)
{
	tmr_t *t, *tnext;
	double tnow;
	now = P_dtime();

	/* P_poll() timing has a precision of 1 ms; if next timer is within that,
	   we should just execute it now; this saves a lot of CPU by not running
	   P_poll() for 0 ms when next timer is within 1 ms */
	tnow = now + 0.0011;

	/* maintain timers if we reached time for the earliest timer */
	if (tnow >= next) {
		next = now + DEFAULT_PERIOD;
		/* check each timer if they need to run */
		for(t = timers; t != NULL; t = tnext) {
			tnext = t->next;
			if ((t->when <= tnow) && (!t->del)) {
				virt_ctrl_t ret;
				assert(t->proc->timer != NULL);
				ret = t->proc->timer(t->proc, t->data);
				vp_act(t->proc, ret);
				timer_del(t, 1);
			}
			else
				update_next(t->when);
		}

		/* revisit each timer and delete ones marked for deletion */
		for(t = timers; t != NULL; t = tnext) {
			tnext = t->next;
			if (t->del)
				timer_del(t, 0);
		}
	}

	/* calculate how much is left until next timer triggers */
#ifdef TIMER_DEBUG
	fprintf(stderr, "timer next: %d\n", (int)((next - now) * 1000.0));
#endif
	return (int)((next - now) * 1000.0);
}

void timer_init(void)
{
	now = P_dtime();
	next = now + DEFAULT_PERIOD;
}

void timer_uninit(void)
{
	tmr_t *t, *tnext;
	for(t = timers; t != NULL; t = tnext) {
		tnext = t->next;
		timer_del(t, 0);
	}
}
